/*
 * Created on 06.09.2005 for PIROL
 *
 * SVN header information:
 *  $Author: LBST-PF-3\orahn $
 *  $Rev: 2519 $
 *  $Date: 2006-10-11 11:53:07 +0200 (Mi, 11 Okt 2006) $
 *  $Id: LayerProjector.java 2519 2006-10-11 09:53:07Z LBST-PF-3\orahn $
 */
package de.fhOsnabrueck.jump.pirol.utilities.projection;

import Jama.Matrix;

import com.vividsolutions.jts.geom.Coordinate;
import com.vividsolutions.jts.geom.Geometry;
import com.vividsolutions.jump.feature.Feature;
import com.vividsolutions.jump.workbench.model.Layer;
import com.vividsolutions.jump.workbench.ui.EditTransaction;
import com.vividsolutions.jump.workbench.ui.ErrorHandler;
import com.vividsolutions.jump.workbench.ui.LayerViewPanel;
import com.vividsolutions.jump.workbench.ui.task.TaskMonitorDialog;

import de.fhOsnabrueck.jump.pirol.utilities.FeatureCollection.PirolFeatureCollection;
import de.fhOsnabrueck.jump.pirol.utilities.apiTools.FeatureCollectionTools;
import de.fhOsnabrueck.jump.pirol.utilities.debugOutput.DebugUserIds;
import de.fhOsnabrueck.jump.pirol.utilities.debugOutput.PersonalLogger;
import de.fhOsnabrueck.jump.pirol.utilities.i18n.PirolPlugInMessages;
import de.fhOsnabrueck.jump.pirol.utilities.metaData.MetaInformationHandler;
import de.fhOsnabrueck.jump.pirol.utilities.projection.Projections.AbstractProjection;
import de.fhOsnabrueck.jump.pirol.utilities.projection.Projections.GaussKruegerProjection;
import de.fhOsnabrueck.jump.pirol.utilities.projection.ellipsoid.AbstractEllipsoid;
import de.fhOsnabrueck.jump.pirol.utilities.projection.ellipsoid.BesselEllipsoid;

/**
 * This class transforms the coordinates of features in a layer from one (WGS84, only yet) to
 * an other projection (Gau-Krger, only yet) and applies the resulting coordinates to the layer
 *
 * @author Ole Rahn
 * <br>
 * <br>FH Osnabr&uuml;ck - University of Applied Sciences Osnabr&uuml;ck,
 * <br>Project: PIROL (2005),
 * <br>Subproject: Daten- und Wissensmanagement
 * 
 * @version $Rev: 2519 $
 * 
 */
public class LayerProjector extends Thread {
    
    protected static PersonalLogger logger = new PersonalLogger(DebugUserIds.OLE);
    
    public final static String[] validInputProjections = new String[]{ AbstractProjection.KEY_WGS84, AbstractProjection.KEY_GAUSS_KRUEGER };
    public final static String[] validOutputProjections = new String[]{ AbstractProjection.KEY_GAUSS_KRUEGER, AbstractProjection.KEY_WGS84 };
    
    protected ReprojectionDescription reprojectionDescription = null;
    
    // constant values
    protected final static double oneBy6 = 1.0 / 6.0;
    protected final static double oneBy24 = 1.0 / 24.0;
    protected final static double oneBy120 = 1.0 / 120.0;
    protected final static double nineBy16 = 9.0 / 16.0;
    protected final static double fifteenBy16 = 15.0 / 16.0;
    protected final static double minusThirtyfiveBy48 = -35.0 / 48.0;
    
    // Jump
    protected ErrorHandler errHandler = null;
    protected PirolFeatureCollection featureCollection = null;
    protected Layer layer = null;
    protected TaskMonitorDialog monitor = null;
    protected boolean disposeMonitorAfterUse = true;
    protected LayerViewPanel layerViewPanel = null;
    
    /**
     * Constructor - This constructor does <b>not</b> use EditTransactions. Use this construtor only
     * if the FeatureCollection you pass to this constructor is not attached to a visible Layer.
     *@param fc FeatureCollection which's projection is to be changed 
     *@param reprojectionDescription
     *@param errHandler
     *@param monitor
     *@param layerViewPanel
     *
     *@see LayerProjector#LayerProjector(Layer, ReprojectionDescription, ErrorHandler, TaskMonitorDialog, LayerViewPanel)
     */
    public LayerProjector(PirolFeatureCollection fc, ReprojectionDescription reprojectionDescription, ErrorHandler errHandler, TaskMonitorDialog monitor, LayerViewPanel layerViewPanel){
        this.featureCollection = fc;
        this.errHandler = errHandler;
        
        this.monitor = monitor;
        this.layerViewPanel = layerViewPanel;
        
        this.reprojectionDescription = reprojectionDescription;
    }
    
    /**
     * Constructor. Passing the layer to the LayerProjector enables the use of so called EditTransactions - 
     * this is the way to go, if you work with visible layers.
     *@param layer layer which's projection is to be changed
     *@param reprojectionDescription
     *@param errHandler
     *@param monitor
     *@param layerViewPanel
     *
     *@see EditTransaction
     */
    public LayerProjector(Layer layer, ReprojectionDescription reprojectionDescription, ErrorHandler errHandler, TaskMonitorDialog monitor, LayerViewPanel layerViewPanel){
        this(MetaInformationHandler.createPirolFeatureCollection(layer.getFeatureCollectionWrapper().getUltimateWrappee()), reprojectionDescription, errHandler, monitor, layerViewPanel);
        this.layer = layer;
        this.layer.setFeatureCollection(this.featureCollection);
    }
    
    public final void changeLayerProjection(){
        Feature[] featArray = FeatureCollectionTools.FeatureCollection2FeatureArray(this.featureCollection);
        Feature feat;
        Coordinate[] coordinates, newCoordinates;
        Coordinate coord, coordTmp = null;
        
        Geometry geom, geomTmp;
        
        if (this.reprojectionDescription.startProjectionEqualsTargetProjection() ){
            if (this.errHandler != null)
                this.errHandler.handleThrowable(new Exception( PirolPlugInMessages.getString("source-projection") + " == " + PirolPlugInMessages.getString("target-projection")));
            else
                throw new IllegalArgumentException( PirolPlugInMessages.getString("source-projection") + " == " + PirolPlugInMessages.getString("target-projection"));
            return;
        }
        
        // to GK
        Matrix T = null, R = null;
        // to WGS84
        Matrix R_trans = null;
        
        if (reprojectionDescription.needHelmertTransformation()){
            GaussKruegerProjection gkProjection = reprojectionDescription.getProjectionWithHelmertParameter(); 
        	Matrix[] tmp = LayerProjector.fillMatricesForHelmertTransformation(gkProjection);
        	R = tmp[0];
        	T = tmp[1];
        	R_trans = tmp[2];
        }
        
        boolean editTransactionUsed = true;
        EditTransaction edtr = null;
        
        if (this.layer != null){
            edtr = new EditTransaction(this.featureCollection.getFeatures(), PirolPlugInMessages.getString("change-projection") + ": " + this.reprojectionDescription.toString(), this.layer, true, false, this.layerViewPanel);
        } else {
            editTransactionUsed = false;
        }
        
        for (int i=0; i<featArray.length; i++){
            if (i%500==0 && this.monitor!=null){
                monitor.report(i, featArray.length, PirolPlugInMessages.getString("done"));
            }
            feat = featArray[i];
            geom = feat.getGeometry();
            
            if (!geom.isValid()){
                logger.printError("Geometry ist not valid: " + geom.toText());
            }
            
            if (editTransactionUsed)
                coordinates = edtr.getGeometry(feat).getCoordinates();
            else
                coordinates = feat.getGeometry().getCoordinates();
            
            geomTmp = (Geometry)geom.clone();
            newCoordinates = geomTmp.getCoordinates();
            
            for (int j=0; j<coordinates.length; j++){
                
                coord = coordinates[j];
                
        
                try {

                	coordTmp = LayerProjector.changeProjectionOfCoordinate(coord, this.reprojectionDescription, R, T, R_trans);
                } catch (RuntimeException e) {
                    if (this.errHandler!=null)
                        this.errHandler.handleThrowable(e);
                    else {
                        logger.printError(e.getMessage());
                        throw e;
                    }
                    break;
                }
                
                newCoordinates[j].x = coordTmp.x;
                newCoordinates[j].y = coordTmp.y;
                newCoordinates[j].z = coordTmp.z;
                
            }
            
            
            if (editTransactionUsed) {
                edtr.setGeometry(feat, geomTmp);
            } else {
                if (geomTmp.isValid()){
                    feat.setGeometry(geomTmp);
//                    geomTmp.geometryChanged();
                } else {
                    throw new RuntimeException("invalid geometries");
                }
            }
        }

        if (editTransactionUsed && edtr.proposedGeometriesValid()){
            logger.printDebug("committing geometries!");
            
            edtr.clearEnvelopeCaches();
            edtr.commit();
            
        } else if (editTransactionUsed){
            if (this.errHandler!=null)
                this.errHandler.handleThrowable(new Exception("invalid geometries"));
            else 
                throw new RuntimeException("invalid geometries");
            logger.printError("invalid geoms!");
        } 

        this.featureCollection.invalidateEnvelope(false);
        
        if (this.layer!=null){
            this.layer.fireAppearanceChanged();
        }

    }
    
    /**
     * 
     * @param gkProjection
     * @param R
     * @param T
     * @param R_trans
     * @return Matrix[ R, T, R_trans]
     */
    protected static Matrix[] fillMatricesForHelmertTransformation(GaussKruegerProjection gkProjection){
    	Matrix R,T,R_trans;
        T = new Matrix(new double[]{gkProjection.getTransX(),gkProjection.getTransY(),gkProjection.getTransZ()}, 3);
        R = new Matrix(new double[][]{new double[]{1.0,gkProjection.getRotZ(),-gkProjection.getRotY()}, new double[]{-gkProjection.getRotZ(), 1.0, gkProjection.getRotX()}, new double[]{gkProjection.getRotY(), -gkProjection.getRotX(), 1.0}});
        R_trans = R.transpose();
        
        return new Matrix[]{R,T,R_trans}; 
    }
    
    
    /**
     * 
     * @param geom2Change geometry to change coordinates of
     * @param reprojectionDescription description containing source and target projection 
     * @return new geometry with changed coordinates or null, if ann error occured
     */
    public static Geometry changeProjectionOfGeometry(Geometry geom2Change, ReprojectionDescription reprojectionDescription){
        Geometry geom = geom2Change;
        Coordinate[] newCoordinates, coordinates;
        Coordinate coord, coordTmp;
        
        if (!geom.isValid()){
            logger.printError("Geometry ist not valid: " + geom.toText());
            return null;
        }
        
        coordinates = geom2Change.getCoordinates();
   	
    	Geometry geomTmp = (Geometry)geom2Change.clone();
        newCoordinates = geomTmp.getCoordinates();
        
        for (int j=0; j<coordinates.length; j++){
            coord = coordinates[j];
    
            try {
            	coordTmp = LayerProjector.changeProjectionOfCoordinate(coord, reprojectionDescription);
            } catch (RuntimeException e) {
                logger.printError(e.getMessage());
                throw e;
            }
            
            newCoordinates[j].x = coordTmp.x;
            newCoordinates[j].y = coordTmp.y;
            newCoordinates[j].z = coordTmp.z;
        }
        return geomTmp;
    }
    
    protected static Coordinate changeProjectionOfCoordinate(Coordinate coord2Change, ReprojectionDescription reprojectionDescription, Matrix R, Matrix T, Matrix R_trans){
        
    	Coordinate coordTmp = (Coordinate)coord2Change.clone();
    	
        if (reprojectionDescription.isWgs84ToGaussKrueger()) {
            coordTmp = LayerProjector.wgsToGaussKrueger(coord2Change, reprojectionDescription, R, T);
        } else if (reprojectionDescription.isGaussKruegerToWgs84()){
            coordTmp = LayerProjector.gaussKruegerToWgs(coord2Change, reprojectionDescription, R_trans, T);
        } else if (reprojectionDescription.isGkMeridianBandChange()){
            coordTmp = LayerProjector.gaussKruegerMeridianBandChange(coord2Change, reprojectionDescription);
        } else {
            logger.printError("don't know what to do - aborting!");
            throw new IllegalArgumentException("unclear projection wishes - aborting");
        }
        
        return coordTmp;
    }
    
    
    /**
     *@param coord coordinate in BLH
     *@return coordinate in XYZ
     */
    protected final static Coordinate blhToXyz(Coordinate coord, AbstractEllipsoid ellipsoid){
        double N = ellipsoid.getA() / Math.sqrt( 1.0 - ellipsoid.getESquare() * Math.pow(Math.sin(coord.y),2d) );
        
        double X = (N + coord.z) * Math.cos(coord.y) * Math.cos(coord.x);
        double Y = (N + coord.z) * Math.cos(coord.y) * Math.sin(coord.x);
        double Z = ((1.0 - ellipsoid.getESquare()) * N + coord.z) * Math.sin(coord.y);
        
        return new Coordinate(X,Y,Z);
    }
    
    /**
     * performs a helmert transformation
     *@param coord coordinate
     *@return transformed coordinate
     */
    protected final static Coordinate helmert(Coordinate coord, GaussKruegerProjection gkProjection, Matrix R, Matrix T){
        Matrix Xs = new Matrix( new double[]{coord.x,coord.y,coord.z} ,3);
        Matrix Xz = T.plus( R.times(gkProjection.getMapscale()).times(Xs) );
        
        return new Coordinate(Xz.get(0,0), Xz.get(1,0), Xz.get(2,0));
    }
    
    /**
     * performs transformation back from helmert coordinates
     *@param coord transformed coordinate
     *@return coordinate
     */
    protected final static Coordinate helmert_reverse(Coordinate coord, GaussKruegerProjection gkProjection, Matrix R_trans, Matrix T){
        Matrix Xs = new Matrix( new double[]{coord.x,coord.y,coord.z} ,1);
        Matrix Xz = R_trans.times( Xs.transpose().minus(T).times(gkProjection.getMapscale_inv()) );
        
        return new Coordinate(Xz.get(0,0), Xz.get(1,0), Xz.get(2,0));
    }
    
    /**
     *@param coord coordinate in XYZ
     *@return coordinate in BLH
     */
    protected final static Coordinate xyzToBlh(Coordinate coord, AbstractEllipsoid ellipsoid){
        double pQuadrat = coord.x*coord.x + coord.y*coord.y;
        double p = Math.sqrt(pQuadrat);
        double phi = Math.atan((coord.z*ellipsoid.getA())/(p*ellipsoid.getB()));
        
        double L = Math.atan(coord.y/coord.x);
        double B = Math.atan( (coord.z + ellipsoid.getESquare2() * ellipsoid.getB() * Math.pow(Math.sin(phi),3))/(p - ellipsoid.getESquare() * ellipsoid.getA() * Math.pow(Math.cos(phi),3)) );
        double N = ellipsoid.getA() / Math.sqrt(1-(ellipsoid.getESquare()*Math.pow(Math.sin(B),2d)));
        double H = p/Math.cos(B) - N;
        
        return new Coordinate(L,B,H);
    }
    
    /**
     * converts BLH coordinates to Gau-Krger conform coordinates
     *@param coord
     *@return GK coordinate
     */
    protected final static Coordinate blhToGk(Coordinate coord, GaussKruegerProjection gkProjection){
        double rVal = 0, hVal = 0;
        BesselEllipsoid besselEllipsoid = (BesselEllipsoid)gkProjection.getEllipsoid();
        
        if (false){
            // Rheinbraun
            // FIXME: Fehler in diesem Zweig der IF-Bedingung
            // -> Nord-Wert vllig ruiniert, Ost-/West-Wert auch falsch
            double t = Math.tan(coord.y);
            double tq = Math.pow(t,2);

            double n_quadrat = besselEllipsoid.getESquare2() * Math.pow(Math.cos(coord.y),2);
            double N_quer = besselEllipsoid.getC() / Math.sqrt(1+n_quadrat);
            double cos_B = Math.cos(coord.y);
            double deltaL = coord.x - gkProjection.getL0();
            
            double p = AbstractEllipsoid.radToDeg;
            
            double param0x = G(coord.y, besselEllipsoid);
            double param2x = 1.0 / (2*p*p) * N_quer * Math.pow(cos_B,2) * t;
            double param4x = 1.0 / (24*Math.pow(p,4)) * N_quer * Math.pow(cos_B,4) * t * (5 - tq + 9 * n_quadrat);
            
            double param1y = 1.0 / p * N_quer * cos_B;
            double param3y = 1.0 / (6*Math.pow(p,3)) * N_quer * Math.pow(cos_B,3) * ( 1 - tq + n_quadrat );
            double param5y = 1.0 / (120*Math.pow(p,5)) * N_quer * Math.pow(cos_B,5) * ( 5 - 18 * tq + Math.pow(t,4) );
            
            double x = param0x + param2x * Math.pow(deltaL,2) + param4x * Math.pow(deltaL,4);
            double y = param1y * deltaL + param3y * Math.pow(deltaL,3) + param5y * Math.pow(deltaL,5);
            
            rVal = gkProjection.getFalseEasting() + y;
            hVal = x;
            
        } else {
            // Labonde
            
            // delta L
            double l = coord.x - gkProjection.getL0();
            
            // cos_b
            double k = Math.cos(coord.y);
            
            // t
            double tt = Math.tan(coord.y);
            double ttq = Math.pow(tt,2);
            
            double eq = besselEllipsoid.getESquare2();
            
            double Vq = 1 + eq * Math.pow(k,2);
            double V = Math.sqrt(Vq);
            double N = Math.pow(besselEllipsoid.getA(), 2)/(besselEllipsoid.getB()*V);
            double n = (besselEllipsoid.getA()-besselEllipsoid.getB())/(besselEllipsoid.getA()+besselEllipsoid.getB());       
            
            double x = 0.5 * N * tt * Math.pow(k,2) * Math.pow(l,2) + oneBy24 * N * tt * (9 * Vq - ttq - 4) * Math.pow(k,4) * Math.pow(l,4);
            double g = coord.y + ( (-1.5 * n + nineBy16 * Math.pow(n,3)) * Math.sin(2*coord.y) + fifteenBy16 * Math.pow(n,2) * Math.sin(4*coord.y) + minusThirtyfiveBy48 * Math.pow(n,3) * Math.sin(6*coord.y) );
            
            // FIXME: kick out this formula and use the other one with the same C for both ways wgs <-> gk!!
//            double S = g * besselEllipsoid.getC();
            double S = g * besselEllipsoid.getC_Labonde();
            
            hVal = S + x;
            
            double y = N * k * l + oneBy6 * N * (Vq - ttq) * Math.pow(k,3) * Math.pow(l,3) + oneBy120 * N * (5 - 18.0 * ttq + Math.pow(tt,4)) * Math.pow(k,5) * Math.pow(l,5);
            rVal = gkProjection.getFalseEasting() + y;            
        }
        
        return new Coordinate(rVal,hVal,coord.z);

    }
    
    /**
     * determine the geografical latitude B from a given Meridian arc length 
     *@param G Meridian arc length
     *@return geografical latitude B (degrees)
     */
    protected final static double B(double G, GaussKruegerProjection gkProjection){
        
        BesselEllipsoid besselEllipsoid = (BesselEllipsoid)gkProjection.getEllipsoid();
        
        double sigma = (G/besselEllipsoid.getParamE0());
        
        double B = sigma + besselEllipsoid.getParamF2() * Math.sin(2.0*sigma*AbstractEllipsoid.degToRad) + besselEllipsoid.getParamF4() * Math.sin(4.0*sigma*AbstractEllipsoid.degToRad) + besselEllipsoid.getParamF6() * Math.sin(6.0*sigma*AbstractEllipsoid.degToRad);
        
        return B;
    }
    
    /**
     * converts Gau-Krger conform coordinates to BLH coordinates (ellipsoid: bessel)
     *@param coord
     *@return BLH coordinate
     */
    protected final static Coordinate gkToBlh(Coordinate coord, GaussKruegerProjection gkProjection){
        double x = coord.y;
        double y = coord.x - gkProjection.getFalseEasting();
        
        BesselEllipsoid besselEllipsoid = (BesselEllipsoid)gkProjection.getEllipsoid();
        
        double Lh = gkProjection.getL0() * AbstractEllipsoid.radToDeg;
        double Bf = LayerProjector.B(x,gkProjection);
        
//        this.logger.printDebug("Lh: " + Lh + ", Bf:" + Bf);
        
        double rad = AbstractEllipsoid.radToDeg;
        
        double c = besselEllipsoid.getC(),
               tf = Math.tan(Bf * AbstractEllipsoid.degToRad), // ok
               tf2 = Math.pow(tf,2.), // ok
               cos_Bf = Math.cos(Bf * AbstractEllipsoid.degToRad), // ok
               nf2 = besselEllipsoid.getESquare2() * Math.pow(cos_Bf,2), // ok
               Nf_quer = c / (Math.sqrt(1 + nf2)); // ok
        
        
        double B2 =  - rad / (2.0 * Math.pow(Nf_quer,2.0)) * tf * ( 1 + nf2 ); // ok
        double B4 =    rad / (24.0 * Math.pow(Nf_quer,4.0)) * tf * ( 5.0 + 3.0 * tf2 + 6.0 * nf2 * ( 1 - tf2) ); // ok
        double B6 =  - rad / (720.0 * Math.pow(Nf_quer,6.0)) * tf * ( 61.0 + 90.0 * tf2 + 45.0 * Math.pow(tf,4) ); // ok
        
        double L1 =    rad / (Nf_quer * cos_Bf); // ok
        double L3 =  - rad / (6.0 * Math.pow(Nf_quer,3.0) * cos_Bf) * ( 1. + 2. * tf2 + nf2 ); // ok
        double L5 =    rad / (120.0 * Math.pow(Nf_quer,5.0) * cos_Bf) * ( 5. + 28. * tf2 + 24. * Math.pow(tf,4) ); // ok
        
        double B = Bf + B2 * Math.pow(y,2) + B4 * Math.pow(y,4) + B6 * Math.pow(y,6); // ok
        double L = Lh + L1 * y + L3 * Math.pow(y,3) + L5 * Math.pow(y,5); // ok
        
        // es wird in den folgenden Schritten mit rad weitergerechnet, das ergebnis ist jedoch in deg
//        logger.printDebug("B: " + B + (B*degToRad));
//        logger.printDebug("L: " + L + (L*degToRad));
        
        return new Coordinate(L*AbstractEllipsoid.degToRad,B*AbstractEllipsoid.degToRad,coord.z);
    }
    
    public static Coordinate changeProjectionOfCoordinate(Coordinate coordinate, ReprojectionDescription reprojectionDescription){
    	// to GK
        Matrix T = null, R = null;
        // to WGS84
        Matrix R_trans = null;
        
    	if (reprojectionDescription.needHelmertTransformation()){
    		GaussKruegerProjection gkp = reprojectionDescription.getProjectionWithHelmertParameter();
    		Matrix[] tmp = LayerProjector.fillMatricesForHelmertTransformation(gkp);
    		
    		R= tmp[0];
    		T= tmp[1];
    		R_trans = tmp[2];
    	}
    	
    	return LayerProjector.changeProjectionOfCoordinate(coordinate, reprojectionDescription, R, T, R_trans);
    }
    
    public static Feature changeFeatureProjection(Feature feature, ReprojectionDescription reprojectionDescription){
    	
    	Geometry newGeom = (Geometry)feature.getGeometry().clone();
    	Coordinate[] origCoords = feature.getGeometry().getCoordinates();
    	Coordinate[] newCoords = newGeom.getCoordinates();
    	
    	Coordinate newCoord = null;
    	
    	for (int i=0; i<origCoords.length; i++){
    		newCoord = LayerProjector.changeProjectionOfCoordinate(origCoords[i], reprojectionDescription);
    		
    		newCoords[i].x = newCoord.x;
    		newCoords[i].y = newCoord.y;
    		newCoords[i].z = newCoord.z;
    	}
    	
    	feature.setGeometry(newGeom);
    	
    	return feature;
    }
    
    protected static void printCoordinate(Coordinate c, double factor){
        logger.printDebug(c.x * factor + ", " + c.y * factor + ", " + c.z * factor);
    }
    
    protected final static Coordinate gaussKruegerToWgs(Coordinate coord, ReprojectionDescription reprojection, Matrix R_trans, Matrix T){
        if (Double.isNaN(coord.z)) coord.z = 0;
        
        GaussKruegerProjection gkProj = (GaussKruegerProjection)reprojection.getStartProjection();
        
        Coordinate blh = LayerProjector.gkToBlh(coord, gkProj);
//        this.printCoordinate(blh, radToDeg);
        
        Coordinate helmertXYZ = LayerProjector.blhToXyz(blh, gkProj.getEllipsoid());
//        this.printCoordinate(helmertXYZ, 1.0 );
        
        Coordinate xyz = helmert_reverse(helmertXYZ,gkProj,R_trans,T);
//        this.printCoordinate(xyz, 1.0);
        
        Coordinate wgs = LayerProjector.xyzToBlh(xyz, reprojection.getTargetProjection().getEllipsoid() );
//        this.printCoordinate(wgs, radToDeg);
        
        wgs.x *= AbstractEllipsoid.radToDeg;
        wgs.y *= AbstractEllipsoid.radToDeg;
        wgs.z *= AbstractEllipsoid.radToDeg;
        
        return wgs;
    }
    
    
    protected final static Coordinate gaussKruegerMeridianBandChange(Coordinate coord, ReprojectionDescription reprojection){
        if (Double.isNaN(coord.z)) coord.z = 0;
        
        Coordinate blh = LayerProjector.gkToBlh(coord, (GaussKruegerProjection)reprojection.getStartProjection());
        return blhToGk(blh, (GaussKruegerProjection)reprojection.getTargetProjection());
    }
    
    protected final static Coordinate wgsToGaussKrueger(Coordinate coord, ReprojectionDescription reprojection, Matrix R, Matrix T){
        double L = coord.x * AbstractEllipsoid.degToRad;
        double B = coord.y * AbstractEllipsoid.degToRad;
        double H = coord.z * AbstractEllipsoid.degToRad;
        
        if (Double.isNaN(H)) H = 0;
        
        Coordinate xyz = LayerProjector.blhToXyz(new Coordinate(L,B,H), reprojection.getStartProjection().getEllipsoid());
        
        Coordinate helmertXYZ = helmert(xyz, (GaussKruegerProjection)reprojection.getTargetProjection(), R, T);
        
        Coordinate blh = LayerProjector.xyzToBlh(helmertXYZ, reprojection.getTargetProjection().getEllipsoid());
        
        return blhToGk(blh, (GaussKruegerProjection)reprojection.getTargetProjection());
    }
    
    protected final static double  G(double B, BesselEllipsoid ellipsoid){
        double E0 = ellipsoid.getParamE0();
        double E2 = ellipsoid.getParamE2();
        double E4 = ellipsoid.getParamE4();
        double E6 = ellipsoid.getParamE6();
        /*
        this.logger.printDebug("E0: " + Double.toString(E0));
        this.logger.printDebug("E2: " + Double.toString(E2));
        this.logger.printDebug("E4: " + Double.toString(E4));
        this.logger.printDebug("E6: " + Double.toString(E6));
        */
            

        return E0 * B + E2 * Math.sin(2*B*AbstractEllipsoid.degToRad) + E4 * Math.sin(4*B*AbstractEllipsoid.degToRad) + E6 * Math.sin(6*B*AbstractEllipsoid.degToRad);
    }

    
    
    public boolean isDisposeMonitorAfterUse() {
        return disposeMonitorAfterUse;
    }

    public void setDisposeMonitorAfterUse(boolean disposeMonitorAfterUse) {
        this.disposeMonitorAfterUse = disposeMonitorAfterUse;
    }

    /**
     * Attention: after the execution of this method (also used by start()), the TaskMonitor will be disposed!
     *
     */
    public void run() {
        try {
            this.changeLayerProjection();
        } catch (RuntimeException e) {
            logger.printError(e.getLocalizedMessage());
            throw e;
        }
        
        if (this.monitor != null && this.disposeMonitorAfterUse){
            this.monitor.setVisible(false);
            this.monitor.dispose();
        }
    }
    
    

}
